__all__ = ['__version__', '__versiondate__', '__license__']

__version__      = '0.17.1'
__versiondate__  = '2022-02-02'
__license__      = 'ScirisWeb %s (%s) -- (c) Sciris.org' % (__version__, __versiondate__)
